@@ -148,6 +148,10 @@ class ValueTraceBase(object):
     def isUnknownTrace():
         return False
 
+    @staticmethod
+    def isUnknownOrVeryTrustedTrace():
+        return False
+
     @staticmethod
     def isEscapeTrace():
         return False
@@ -272,6 +276,9 @@ class ValueTraceUninitialized(ValueTraceUnassignedBase):
     def isTraceThatNeedsEscape():
         return False
 
+    def dump(self, indent):
+        print("%s%s %s:" % (indent, self.__class__.__name__, id(self)))
+
 
 class ValueTraceDeleted(ValueTraceUnassignedBase):
     """Trace caused by a deletion."""
@@ -382,6 +389,10 @@ class ValueTraceUnknown(ValueTraceBase):
     def isUnknownTrace():
         return True
 
+    @staticmethod
+    def isUnknownOrVeryTrustedTrace():
+        return True
+
     @staticmethod
     def isTraceThatNeedsEscape():
         return False
@@ -441,6 +452,10 @@ class ValueTraceEscaped(ValueTraceUnknown):
     def isUnknownTrace():
         return False
 
+    @staticmethod
+    def isUnknownOrVeryTrustedTrace():
+        return False
+
     @staticmethod
     def isEscapeTrace():
         return True
@@ -525,11 +540,9 @@ class ValueTraceAssign(ValueTraceBase):
             return None
 
     def getAttributeNodeVeryTrusted(self):
-        source_node = self.assign_node.subnode_source
-
         # Hard imports typically.
-        if source_node.hasVeryTrustedValue():
-            return source_node
+        if self.assign_node.hasVeryTrustedValue():
+            return self.assign_node.subnode_source
         else:
             return None
 
@@ -543,6 +556,12 @@ class ValueTraceAssignUnescapable(ValueTraceAssign):
         return False
 
 
+class ValueTraceAssignVeryTrusted(ValueTraceAssignUnescapable):
+    @staticmethod
+    def isUnknownOrVeryTrustedTrace():
+        return True
+
+
 class ValueTraceAssignUnescapablePropagated(ValueTraceAssignUnescapable):
     """Assignment from value where it is not that escaping doesn't matter."""
 
@@ -584,6 +603,12 @@ class ValueTraceMergeBase(ValueTraceBase):
         self.addUsage()
         self.merge_usage_count += 1
 
+    def dump(self, indent):
+        print("%s%s %s:" % (indent, self.__class__.__name__, id(self)))
+
+        for trace in self.previous:
+            trace.dump(indent + "  ")
+
 
 class ValueTraceMerge(ValueTraceMergeBase):
     """Merge of two or more traces.
@@ -596,6 +621,23 @@ class ValueTraceMerge(ValueTraceMergeBase):
     __slots__ = ()
 
     def __init__(self, traces):
+        shorted = []
+
+        for trace in traces:
+            if type(trace) is ValueTraceMerge:
+                for trace2 in trace.previous:
+                    if trace2 not in shorted:
+                        shorted.append(trace2)
+            else:
+                if trace not in shorted:
+                    shorted.append(trace)
+
+        traces = tuple(shorted)
+
+        assert len(traces) > 1
+
+        # assert len(set(traces)) == len(traces), [(v) for v in traces]
+
         ValueTraceMergeBase.__init__(self, owner=traces[0].owner, previous=traces)
 
     def __repr__(self):
